#include <cstdio>
#include <stdint.h>
#include <iostream>
#include <cstring>
#include "test.h"

constexpr uint32_t BIT_FLOAT_1 = 0x0;
constexpr uint32_t BIT_FLOAT_NEG_1 = 0x80000000;
#define BIT_FLOAT_MUL(x, y) ((x) ^ (y))
#define EPSILON 1e-6

struct Point2D
{
	float x;
	float y;
};

struct Line
{
	void update(const Point2D *__restrict__ A,
				const Point2D *__restrict__ B)
	{
		a = B->y - A->y;
		b = A->x - B->x;
		c = a * (A->x) + b * (A->y);
		// printf("c:%f \n", c);
		// printf("a=B->y - A->y:%f=%f-%f \n", a, B->y, A->y);
	}
	float a; // (y1 - y0)
	float b; // (x0 - x1)
	float c; // (x0y1 -x1y0) 叉积
};

static uint32_t getDirection(const Point2D *__restrict__ A)
{
	// 叉积判断顺时针 或 逆时针
	float x0 = A[2].x - A[0].x; // w
	float y0 = A[2].y - A[0].y; // h
	float x1 = A[3].x - A[1].x; // w
	float y1 = A[3].y - A[1].y; // -h
	// printf("--------------[x0, y0, x1, y1]:[%f, %f, %f, %f],  %f=%f-%f\n", x0, y0, x1, y1, (x0 * y1 - y0 * x1), x0*y1, y0*x1);
	// for(int i=0; i<4; i++){
    // 	printf("--------------A[%d]:[%f, %f]\n", i, A[i].x, A[i].y);
  	// }

	if ((x0 * y1 - y0 * x1) < 0){
		return BIT_FLOAT_1;
	}else{
		return BIT_FLOAT_NEG_1;
	}
}

union FP32U32
{
	float fp32;
	uint32_t u32;
};

static bool isInner(const Line *__restrict__ line,
					const Point2D *__restrict__ c,
					uint32_t direction)
{

	FP32U32 eps;
	eps.fp32 = EPSILON;
	eps.u32 = BIT_FLOAT_MUL(eps.u32, direction);
	
	FP32U32 result;
	result.fp32 = (line->b * c->y + line->a * c->x - line->c + eps.fp32);
	result.u32 = BIT_FLOAT_MUL(result.u32, direction);
	
	printf("------------------>result.fp32:%f  result.u32:%x, direction:%x \n", result.fp32, result.u32, direction);

	return result.fp32 > 0;
}

struct QuadClipBox
{
	void addLines(const Point2D *__restrict__ A)
	{
		direction = getDirection(A);
		printf("--------------getDirection(A):%x \n", direction);
		line[0].update(A, A + 1);

		// for(int i=0; i<4; i++){
		// 	printf("--------------A[%d]:[%f, %f]\n", i, A[i].x, A[i].y);
		// }

		Point2D centerAC = {(A[0].x + A[2].x)/2, (A[0].y + A[2].y)/2};
		bool ACFine = false;
		if (isInner(&line[0], &centerAC, direction))
		{
			line[5].update(A + 3, A);
			if (isInner(&line[5], &centerAC, direction))
			{
				ACFine = true;
			}
			Point2D centerBD = {(A[1].x + A[3].x) / 2, (A[1].y + A[3].y) / 2};
			line[1].update(A + 1, A + 2);
			if (ACFine && isInner(&line[0], &centerBD, direction) && isInner(&line[1], &centerBD, direction))
			{
				is_convex = true;
				return noSplit(A);
			}
		}
		printf("ACFine:%d\n", ACFine);
		if (ACFine)
		{
			splitByAC(A);
		}
		else
		{
			splitByBD(A);
		}
	}

	void splitByAC(const Point2D *__restrict__ A)
	{
		line[2].update(A + 2, A);
		line[3].update(A, A + 2);
		line[4].update(A + 2, A + 3);
	}

	void noSplit(const Point2D *__restrict__ A)
	{
		line[2].update(A + 2, A + 3);
		line[3].update(A + 3, A);
	}

	void splitByBD(const Point2D *__restrict__ A)
	{
		line[1].update(A + 1, A + 3);
		line[2].update(A + 3, A);
		line[3].update(A + 1, A + 2);
		line[4].update(A + 2, A + 3);
		line[5].update(A + 3, A + 1);
	}

	Line line[6];
	uint32_t direction;
	bool is_convex = false;
};

static void cross(const Line *__restrict__ line,
				  const Point2D *__restrict__ c,
				  const Point2D *__restrict__ d,
				  float *__restrict__ o_x,
				  float *__restrict__ o_y)
{
	float a1 = line->a;
	float b1 = line->b;
	float c1 = line->c;
	float lambda = (a1 * c->x + b1 * c->y - c1);
	float beta = (a1 * d->x + b1 * d->y - c1);
	float ratio = lambda / (lambda - beta);
	float m_ratio = 1 - ratio;
	*o_x = ratio * d->x + m_ratio * c->x;
	*o_y = ratio * d->y + m_ratio * c->y;
}

static float polyArea(const Point2D *__restrict__ points, int n)
{

	const Point2D *p0 = points;
	float area = 0;
	float x0 = points[1].x - p0->x;
	float y0 = points[1].y - p0->y;
	for (int i = 1; i < (n - 1); ++i)
	{
		float x1 = points[i + 1].x - p0->x;
		float y1 = points[i + 1].y - p0->y;
		area += (x0 * y1 - y0 * x1);
		x0 = x1;
		y0 = y1;
	}
	area = area / 2;
	printf("------polyArea------>:%f \n", area);
	FP32U32 result;
	result.fp32 = area;
	// result.u32 = (result.u32 & 0x7FFFFFFF);
	return result.fp32;
}

template <int CUTLINE_N>
static float clipArea(const float *__restrict__ box_i,
					  const Line *__restrict__ clip_box_lines,
					  uint32_t direction)
{
	constexpr int MAX_POINT = CUTLINE_N + 5;
	Point2D p_swap0[MAX_POINT];
	Point2D p_swap1[MAX_POINT];
	Point2D *p = p_swap0;
	Point2D *p_next = p_swap1;
#pragma unroll 4
	for (int i = 0; i < 4; ++i)
	{
		p_next[i].x = *box_i;
		++box_i;
		p_next[i].y = *box_i;
		++box_i;
	}
	int n = 4;

	printf("CUTLINE_N:%d \n", CUTLINE_N);
#pragma unroll CUTLINE_N
	for (int i = 0; i < CUTLINE_N; ++i)
	{
		Point2D *tmp = p_next;
		p_next = p;
		p = tmp;

		int new_n = 0;
		const Line *cut_line = &clip_box_lines[i];
		bool prev_inner = isInner(cut_line, p+n-1, direction);
		for (int j = 0; j < n; ++j)
		{
			Point2D *current_point = p + j;
			Point2D *prev_point = p + (j - 1 + n) % n;

			bool current_inner = isInner(cut_line, current_point, direction);
			printf("------------>current_inner:%d, prev_inner:%d \n", current_inner, prev_inner);
			if (current_inner)
			{
				if (!prev_inner)
				{
					cross(cut_line, prev_point, current_point, &p_next[new_n].x, &p_next[new_n].y);
					++new_n;
				}
				p_next[new_n].x = current_point->x;
				p_next[new_n].y = current_point->y;
				++new_n;
			}
			else if (prev_inner)
			{
				cross(cut_line, prev_point, current_point, &p_next[new_n].x, &p_next[new_n].y);
				++new_n;
			}
			prev_inner = current_inner;
		}
		n = new_n;
		printf("----n-------->:%d \n", new_n);

		if (n < 3)
		{
			return 0;
		}
	}
	return polyArea(p_next, n);
}

float intersectArea(const float *__restrict__ box_i,
					const QuadClipBox *__restrict__ clip_box)
{
	float area = 0;
	if (clip_box->is_convex)
	{
		// 凸多边形
		area = clipArea<4>(box_i, clip_box->line, clip_box->direction);
		printf("area route1:%f \n", area);
	}
	else
	{
		// 凹多边形
		area = clipArea<3>(box_i, clip_box->line, clip_box->direction);
		area += clipArea<3>(box_i, &clip_box->line[3], clip_box->direction);
		printf("area route2:%f \n", area);
	}

	return area;
}

static float polyIou(const QuadClipBox *__restrict__ clipbox,
  const float *__restrict__ box_j) {

	float intersect = intersectArea(box_j, clipbox);
	printf("intersect:%f \n", intersect);

	return intersect;
}

// mlu入口函数
int main()
{
	//demo1 相交样例
	float Bbox_A[8] = {0.807617, 0.589844, 0.765137, 0.668457, 0.596680, 0.575684, 0.639160, 0.497559};            
  float Bbox_B[8] = {0.804688, 0.590332, 0.765137, 0.667969, 0.593750, 0.580566, 0.633301, 0.503418};

	//demo2 不相交样例
  // float Bbox_A[8] = {0.807617, 0.589844, 0.765137, 0.668457, 0.596680, 0.575684, 0.639160, 0.497559};            
  // float Bbox_B[8] = {0.435303, 0.691895, 0.397705, 0.764160, 0.226318, 0.674805, 0.263916, 0.602051};

	QuadClipBox clipbox;
	clipbox.addLines(reinterpret_cast<const Point2D *>(&Bbox_A));

	float inter = polyIou(&clipbox, Bbox_B);

	float aera_A = polyArea(reinterpret_cast<const Point2D *>(&Bbox_A), 4);
	float aera_B = polyArea(reinterpret_cast<const Point2D *>(&Bbox_B), 4);

	printf("-------------------IOU------->:%f\n", inter/(aera_A + aera_B - inter + (float)0.000001));

	return 0;
}
